//---------------------------------------------------------------------------------
// Microchip 25AA256 sample
//
//  EEPROM   arduino pin No
// --------  --------------
// 1 CS       Analog 0
// 2 SO       digital 12 MISO
// 3 WP       5V
// 4 GND      GND
// 5 SI       digital 11 MOSI
// 6 SCK      digital 13 SCK
// 7 HOLD     5V
// 8 VCC      5V
//------------------- copyleft iizukagiken 2014 all rights reversed ----------------
#define EEPROM_CS A0
#define EEPROM_MAX_ADDR_LENGTH 32768 // (32kbyte=256kbit)
#include <SPI.h>

typedef union {
  word word_val;
  struct {
    byte L;
    byte H;
  } byte_val;    
} type_convert_word_to_byte; 

//--------------------------------------------------------------------------------
// ２バイトデータを書き込む
//　　引数１はアドレスではなく何ワード目に書き込むかを表すインデックス
//
//  word_index: 0,1,2 ....
//--------------------------------------------------------------------------------
void spi_eeprom_write_word( word word_index, word data )
{
  type_convert_word_to_byte udata;
  type_convert_word_to_byte uaddr;

  uaddr.word_val = (word_index << 1);
  udata.word_val = data;
  spi_eeprom_write_enable();
  digitalWrite( EEPROM_CS, LOW );    // eeprom chip select enable 
  SPI.transfer( B00000010 );         // eeprom Write Instruction
  SPI.transfer( uaddr.byte_val.H );  // eeprom 16bit address 
  SPI.transfer( uaddr.byte_val.L );
  SPI.transfer( udata.byte_val.L );  // write word data
  SPI.transfer( udata.byte_val.H );
  digitalWrite( EEPROM_CS, HIGH );  // eeprom chip select disable 
  delay(6);
}

//--------------------------------------------------------------------------------
// 引数１のアドレスから１セクタ（６４バイト）を引数２のデータで生める
//--------------------------------------------------------------------------------
void spi_eeprom_fill( word sector_number, byte data )
{
  type_convert_word_to_byte uaddr;
  
  uaddr.word_val = (sector_number << 6);
  spi_eeprom_write_enable();
  digitalWrite( EEPROM_CS, LOW );    // eeprom chip select enable 
  SPI.transfer( B00000010 );         // eeprom Write Instruction
  SPI.transfer( uaddr.byte_val.H );  // eeprom 16bit address 
  SPI.transfer( uaddr.byte_val.L );
  for( byte i = 0 ; i < 64 ; i++ ) {
    SPI.transfer( data );
  }
  digitalWrite( EEPROM_CS, HIGH );  // eeprom chip select disable 
  delay(6);
}

//--------------------------------------------------------------------------------
// 全メモリーを引数２のデータで生める
//--------------------------------------------------------------------------------
void spi_eeprom_fill_all( byte data )
{
  for( word sector = 0 ; sector < 512 ; sector++ ) {
    spi_eeprom_fill( sector, data );
  }
}

//--------------------------------------------------------------------------------
// 指定のアドレスの内容を１バイト読み出す
//--------------------------------------------------------------------------------
byte spi_eeprom_read_byte( word addr )
{
  type_convert_word_to_byte uaddr;
  byte data;
  
  uaddr.word_val = addr;
  digitalWrite( EEPROM_CS, LOW );   // eeprom chip select enable 
  SPI.transfer( B00000011 );        // eeprom Read Instruction
  SPI.transfer( uaddr.byte_val.H );  // eeprom 16bit address 
  SPI.transfer( uaddr.byte_val.L );
  data = SPI.transfer(0);
  digitalWrite( EEPROM_CS, HIGH );  // eeprom chip select disable 

  return data;
}

//--------------------------------------------------------------------------------
// ２バイトデータを読み出す
//　　引数はアドレスではなく何ワード目を読み出すかを表すインデックス
//
//  word_index: 0,1,2 ....
//--------------------------------------------------------------------------------
word spi_eeprom_read_word( word word_index )
{
  type_convert_word_to_byte udata;
  word addr;

  addr = (word_index << 1);
  udata.byte_val.L = spi_eeprom_read_byte( addr );
  udata.byte_val.H = spi_eeprom_read_byte( addr + 1 );
  
  return udata.word_val;
}

//--------------------------------------------------------------------------------
// 書き込み許可
//--------------------------------------------------------------------------------
void spi_eeprom_write_enable( void )
{
  digitalWrite( EEPROM_CS, LOW  );  // eeprom chip select enable 
  SPI.transfer( B00000110 );        // eeprom enable write operation Instruction
  digitalWrite( EEPROM_CS, HIGH );  // eeprom chip select disable   
}

//--------------------------------------------------------------------------------
// 指定のインデックス内から引数３と一致する一番小さいインデックスを返す
// 範囲外のデータが返されたら見つからなかったという意味
//--------------------------------------------------------------------------------
word spi_eeprom_find_index( word start_index, word end_index, word fill_data )
{
  word offset_index;
  word index;
  word data;
  offset_index = (EEPROM_MAX_ADDR_LENGTH >> 2);

  index = offset_index;
  while( offset_index > 0 ) {  //　バイナリサーチアルゴリズム
    offset_index >>= 1;
    if( index > end_index ) {
      index -= offset_index;
    } else if( index < start_index ) {
      index += offset_index;
    } else {
      data = spi_eeprom_read_word( index );
      if( data == fill_data ) {
        index -= offset_index;
      } else {
        index += offset_index;
      }
    }
  }
  if( spi_eeprom_read_word( index ) != fill_data ) {
    index++;
  }
  
  return index;
}

//--------------------------------------------------------------------------------
// EEPROM初期設定
//--------------------------------------------------------------------------------
void spi_eeprom_init( byte pin_no )
{
  pinMode( pin_no, OUTPUT );
  digitalWrite( pin_no, HIGH );  // eeprom chip select disable 
}

//--------------------------------------------------------------------------------
// arduino setup
//--------------------------------------------------------------------------------
void setup()
{
  byte data;
  word wdata;
  word find_index;
  
  SPI.begin();
  spi_eeprom_init( EEPROM_CS );
  Serial.begin(115200);

  // メモリー内全てを0xFFで埋める　これは２～３秒かかる
  Serial.println( "fill start" );
  spi_eeprom_fill_all( 0xFF );
  Serial.println( "fill end" );

  // １００ワード目～３００ワード目までを2000で埋める
  for( word index = 100 ; index <= 300 ; index++ ) {
    spi_eeprom_write_word( index, 2000 );
  }
  
  // ５０ワード目～１０００ワード目までの中からデータが書き込まれていない（＝値が0xFFFF）先頭インデックスを探す
  // ３００ワード目まで2000で埋められているので301になる
  find_index = spi_eeprom_find_index( 50, 1000, 0xFFFF );
  Serial.print( "index=" );
  Serial.println( find_index );
 
  while(true) {}
}

//--------------------------------------------------------------------------------
// arduino loop
//--------------------------------------------------------------------------------
void loop()
{
  // nothing to do
}
//------------------------------------------------------------- E N D ------------

